/* nl-liststr.c --- newLISP primitives handling lists and strings
//
// Copyright (C) 1992-2006 Lutz Mueller <lutz@nuevatec.com>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2, 1991,
// as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
*/

#include "newlisp.h"
#include "pcre.h"
#include "protos.h"

extern CELL * firstFreeCell;
extern UINT cellCount;
extern CELL * lastCellCopied;
extern SYMBOL * sysSymbol[];


UINT adjustNegativeIndex(int index, UINT length)
{
if(index < 0) index = length + index;
else if((index+1) > length) index = length -1;

return(index < 0 ? 0 : index);
}

#ifdef LINUX
extern char * strcasestr(char * haystack, char * needle);
#endif
CELL * p_member(CELL * params)
{
CELL * key;
CELL * list;
CELL * member;
int caseFlag = TRUE;
char * pos;

key = evaluateExpression(params);

params = params->next;
list = evaluateExpression(params);

if(params->next != nilCell)
	caseFlag = (evaluateExpression(params->next)->type != CELL_NIL);

if(isList(list->type))
	list = (CELL *)list->contents;
else if (list->type == CELL_STRING)
	{
	if(key->type != CELL_STRING)
		return(errorProcExt(ERR_STRING_EXPECTED, params));
#ifndef MINGW
	if(caseFlag)
		pos = strstr((char *)list->contents, (char *) key->contents);
	else
#ifndef SOLARIS
		pos = strcasestr((char *)list->contents, (char *) key->contents);
#else
		pos = strstr((char *)list->contents, (char *) key->contents);
#endif
#else
	pos = strstr((char *)list->contents, (char *) key->contents);
#endif
		
	if(pos) return(stuffString(pos));
	else return(nilCell);	
	}
else 
	return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params->next));

while(list != nilCell)
	{
	if(compareCells(key, list) == 0) break;
	list = list->next;
	}

if(list == nilCell) return(nilCell);
member = getCell(CELL_EXPRESSION);
member->contents = (UINT)copyList(list);
return(member);
}

CELL * p_length(CELL * params)
{
UINT length;
SYMBOL * symbol;

params = evaluateExpression(params);
length = 0;
switch(params->type)
	{
	case CELL_INT32:
		length = sizeof(UINT); break;
	case CELL_FLOAT:
		length = sizeof(double); break;
	case CELL_STRING:
		length = params->aux - 1; break;
	case CELL_CONTEXT:
	case CELL_SYMBOL:
		symbol = (SYMBOL *)params->contents;
		length = strlen(symbol->name);
		break;
	case CELL_DYN_SYMBOL:
		length = strlen((char *)params->contents);
		break;
	case CELL_EXPRESSION:
	case CELL_LAMBDA:
	case CELL_MACRO:
	        length = listlen((CELL *)params->contents);
		break;
	case CELL_ARRAY:
		length = (params->aux - 1) / sizeof(UINT);
	default:
		break;
	}
return(stuffInteger(length));
}


CELL * p_append(CELL * params)
{
CELL * list = NULL; 
CELL * firstCell;
CELL * copy;
CELL * cell;

firstCell = NULL;

while(params != nilCell)
    {
    cell = evaluateExpression(params);
    if(!isList(cell->type))
        {
        if(list == NULL && cell->type == CELL_STRING)
            return(append(cell, params, "", 0, TRUE));

        return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params));
        }

    if(list == NULL)
        list = getCell(cell->type);

    copy = copyList((CELL *)cell->contents);

    params = params->next;
    if(copy == nilCell) continue;

    if(firstCell == NULL) list->contents = (UINT)copy;
    else firstCell->next = copy;

    firstCell = lastCellCopied;
    }

if(list == NULL) 
	return(getCell(CELL_EXPRESSION));

return(list);
}


CELL * append(CELL * cell, CELL * list, char * joint, int jointLen, int evalFlag)
{ 
char * string;
CELL * result;
char * * strPtr;
int * intPtr;
UINT length, N, i, len;

length = 0;

if((N = listlen(list)) == 0) return(stuffString(""));

strPtr = (char * *)allocMemory(N * sizeof(char *));
intPtr = (int *)allocMemory(N * sizeof(int *));
for(i = 0; i < N; i++)
	{
	if(i == 0 && cell != NULL)
		{
		string = (char *)cell->contents;
		len = cell->aux - 1;
		list = list->next;
		}	
	else list = getStringSize(list, &string, (UINT*)&len, evalFlag);
	length += len;
        intPtr[i] = len;
	strPtr[i] = string;
	}

result = getCell(CELL_STRING);
result->aux = length + (N - 1) * jointLen + 1;
result->contents = (UINT)callocMemory((UINT)result->aux);

length = 0;
for(i = 0; i < N; i++)
	{
	if(jointLen > 0 && i > 0) 
		{
		memcpy((char *)result->contents + length, joint, jointLen);
		length += jointLen;
		}
	memcpy((char *)result->contents + length, strPtr[i], intPtr[i]);
	length += intPtr[i];
	}

free(strPtr);
free(intPtr);
return(result);
}


CELL * p_chop(CELL * params)
{
UINT number = 1;
UINT length = 0;
CELL * next;
#ifdef SUPPORT_UTF8
char * ptr;
#endif

next = params->next;
params = evaluateExpression(params);

if(next != nilCell)
	getInteger(next, &number);

if(params->type == CELL_STRING)
	{
#ifndef SUPPORT_UTF8
	length = params->aux - 1;
	if(number > length) number = length;
	length = length - number;
	return stuffStringN((char *)params->contents, length);
#else
	length = utf8_wlen((char *)params->contents);
	if(number > length) number = length;
	length = length - number;
	ptr = (char *)params->contents;
	while(length--)
		ptr += utf8_1st_len(ptr);
	return stuffStringN((char *)params->contents, ptr - (char *)params->contents);
#endif
	}

if(!isList(params->type))
    return(errorProc(ERR_LIST_OR_STRING_EXPECTED));	

length = listlen((CELL *)params->contents);
if(number > length) number = length;

return(sublist((CELL *)params->contents, 0, length - number));
}


/* multi dimensional nth */
CELL * p_nth(CELL * params)
{
int index[16];
int p = 0;
int i;
#ifndef SUPPORT_UTF8
char str[2];
#endif
char * ptr;
CELL * list = nilCell;

while(params != nilCell && p < 16)
	{
	list = evaluateExpression(params);
	if(list->type == CELL_INT32 || list->type == CELL_FLOAT)
		getIntegerExt(list, (UINT*)&index[p++], FALSE);
	else if(list->type == CELL_ARRAY) return(nthArray(index, p, list));
	else if(list->type == CELL_STRING)
		{
		ptr = (char *)list->contents;
#ifndef SUPPORT_UTF8
		index[0] = adjustNegativeIndex(index[0], list->aux - 1);
		str[0] = *(ptr + index[0]);
		str[1] = 0;
		return(stuffString(str));
#else
		index[0] = adjustNegativeIndex(index[0], utf8_wlen(ptr));
		for(i = 0; i < index[0]; i++)
			{
			p = utf8_1st_len(ptr);
			ptr += p;
			}
		return(stuffStringN(ptr, utf8_1st_len(ptr)));
#endif
		}
        else if(p == 0) return(errorProcExt(ERR_NUMBER_EXPECTED, params));
	else if(!isList(list->type))
		return(errorProcExt(ERR_ARRAY_LIST_OR_STRING_EXPECTED, params));
	params = params->next;
	} 
		
/* iterate in and into list(s) */
i = 0;
while(isList(list->type))
    {
    list = (CELL *)list->contents;

    if(index[i] < 0) index[i] = convertNegativeOffset(index[i], list);

    while(index[i]--) 
        {
        if(list->next == nilCell) break;
        list = list->next;
        }

    /* if indices are exhausted or current element is not a list return */
    i++;
    if(i == p || !isList(list->type)) return(copyCell(list));
    }
    
if(p == 0)
	return(errorProc(ERR_MISSING_ARGUMENT));
return(errorProcExt(ERR_ARRAY_LIST_OR_STRING_EXPECTED, params));
}

CELL * setNthStr(CELL * cellStr, CELL * new, int index, int typeFlag);
CELL * setNth(CELL * params, int typeFlag);

CELL * p_setNth(CELL * params) {return setNth(params, 0);}
CELL * p_nthSet(CELL * params) {return setNth(params, 1);}

CELL * setNth(CELL * params, int typeFlag)
{
int index[16];
int p = 0;
CELL * list = nilCell; 
CELL * next;
SYMBOL * sPtr = NULL;

/* new syntax, distinguished by type of first arg and number of args */
next = params->next;
if(params->type == CELL_EXPRESSION && next->next == nilCell )
	{
	list = (CELL *)params->contents;
	params = list->next;	
	if(isSymbol(list->type))
		{
		if(list->type == CELL_SYMBOL)
			sPtr = (SYMBOL *)list->contents;
		else 
			sPtr =	getDynamicSymbol(list);

		list = (CELL *)sPtr->contents;

		if(list->type == CELL_CONTEXT)
			{
			list = (CELL *)(translateCreateSymbol(
				((SYMBOL*)list->contents)->name,
				CELL_NIL,
				(SYMBOL*)list->contents,
				TRUE))->contents;
			}
		else if(isProtected(sPtr->flags))
				return(errorProcExt2(ERR_SYMBOL_PROTECTED, stuffSymbol(sPtr)));
		}
	else
		list = evaluateExpression(list);

	while(params != nilCell)
		params = getInteger(params, (UINT*)&index[p++]);

	if(list->type == CELL_EXPRESSION)
		return(setNthList(index, p, list, next, typeFlag));
	if(list->type == CELL_STRING)
		return(setNthStr(list, next, index[0], typeFlag));
	if(list->type == CELL_ARRAY)
		return(setNthArray(index, p, list, next, typeFlag));

	return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, list));
	}

/* old syntax */
params = getInteger(params, (UINT*)&index[p++]);

while(params != nilCell && p < 16)
	{
	if(isSymbol(params->type))
		{
		if(params->type == CELL_SYMBOL)
			sPtr = (SYMBOL *)params->contents;
		else
			sPtr = getDynamicSymbol(params);

    	if(isProtected(sPtr->flags))
			{
			/* make sure this is not an index variable */
			if(isEnvelope(symbolType(sPtr)) || symbolType(sPtr) == CELL_STRING)
				return(errorProcExt(ERR_SYMBOL_PROTECTED, params));
			}

		list = (CELL *)sPtr->contents;
		}
	else list = evaluateExpression(params);

	if(list->type == CELL_INT32 || list->type == CELL_FLOAT)
		getIntegerExt(list, (UINT*)&index[p++], FALSE);

	else if(list->type == CELL_ARRAY) return(setNthArray(index, p, list, params->next, typeFlag));
	else if(list->type == CELL_STRING)  return(setNthStr(list, params->next, index[0], typeFlag));
	else if(isList(list->type)) break;
	else 
		{
		if(p == 0)
			return(errorProcExt(ERR_NUMBER_EXPECTED, params));	
		return(errorProcExt(ERR_ARRAY_LIST_OR_STRING_EXPECTED, params));
		}
	params = params->next;
	} 

return(setNthList(index, p, list, params->next, typeFlag));
}


CELL * setNthList(int * index, int p, CELL * list, CELL * params, int typeFlag)
{
CELL * new;  /* un-evaluated replacement */
CELL * org;  /* the original list */
CELL * prev = nilCell;
int i = 0;

org = list;
while(isList(list->type))
    {
    list->aux = (UINT)nilCell; /* undo push optimization */
    list = (CELL *)list->contents;

    if(index[i] < 0) index[i] = convertNegativeOffset(index[i], list);

    while(index[i]--) 
        {
        if(list->next == nilCell) break;
        list = list->next;
        }

    /* if indices are exhausted or element is not list, do replacement */
    i++;
    if(i == p || !isList(list->type)) 
        {
		deleteList((CELL*)sysSymbol[0]->contents);
		sysSymbol[0]->contents = (UINT)copyCell(list);
		new = copyCell(evaluateExpression(params)); 

		/* save previous content */
		prev = getCell(list->type);
		prev->aux = list->aux;
		prev->contents = list->contents;
		prev->next = nilCell;

		list->type = new->type;
		list->aux = new->aux;
		list->contents = new->contents;

		/* free the cell */ 
		new->type = CELL_FREE;
		new->aux = 0;
		new->contents = 0;
		new->next = firstFreeCell;
		firstFreeCell = new;
		--cellCount;

		if(typeFlag) /* nth-set */
			return(prev);
		else         /* set-nth */
			{
			deleteList(prev);
			return(copyCell(org));
			}
        }

    }

if(p == 0)
	return(errorProc(ERR_MISSING_ARGUMENT));

return(errorProcExt(ERR_ARRAY_LIST_OR_STRING_EXPECTED, list));
}


#define INSERT_BEFORE 0
#define INSERT_AFTER 1
#define INSERT_END 2
CELL * p_push(CELL * params)
{
CELL * newCell;
CELL * list;
CELL * cell = NULL;
SYMBOL * sPtr;
int index, insert = 0, evalFlag = 0;

newCell = evaluateExpression(params);
params = params->next;

if(isSymbol(params->type))
    {
    if(params->type == CELL_SYMBOL)
    	sPtr = (SYMBOL *)params->contents;
    else
        sPtr = getDynamicSymbol(params);

    if(isProtected(sPtr->flags))
		return(errorProcExt(ERR_SYMBOL_PROTECTED, params));

    if(!isList(((CELL*)sPtr->contents)->type))
        {
		if(((CELL *)sPtr->contents)->contents == (UINT)nilCell 
           || ((CELL *)sPtr->contents)->contents == (UINT)nilSymbol)
			{
        	deleteList((CELL*)sPtr->contents);
        	list = getCell(CELL_EXPRESSION);
        	sPtr->contents = (UINT)list;
			}
        }
    list = (CELL*)sPtr->contents;
    }
else
    list = evalCheckProtected(params, NULL);

if(!isList(list->type))
	{
	if(list->type == CELL_STRING)
		return(pushOnString(newCell, list, params->next));
	else
		return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params));
	}

if(params->next == nilCell) 
	{
	params = params->next;
	index = 0;
	}
else 
	{
	cell = ((CELL*)params->next)->next;
	params = evaluateExpression(params->next);
	if(isList(params->type))
		{
		evalFlag = FALSE;
		params = getIntegerExt((CELL*)params->contents, (UINT*)&index, FALSE);
		}
	else 
		{
		evalFlag = TRUE;
		getIntegerExt(params, (UINT*)&index, FALSE);
		params = cell;
		}
	}

if(index == -1) 
	{
/*	if(isList(list->type) && params == nilCell) */
	if(params == nilCell)
		{
		newCell = copyCell(newCell);
		cell = (CELL*)list->aux;	
		list->aux = (UINT)newCell;
		if(cell != nilCell)
			{
			cell->next = newCell;
			return(copyCell(newCell));
			}

		if(list->contents == (UINT)nilCell)
			{
			list->contents = (UINT)newCell;
			return(copyCell(newCell));
			}

		list = (CELL *)list->contents;
		while(list->next != nilCell)
			list = list->next;
		list->next = newCell;
		return(copyCell(newCell));
		}

	index = 0x7fffffff;
	}
	
list->aux = (UINT)nilCell; /* undo push optimization */

while(isList(list->type))
    {
    cell = list;
    list = (CELL *)list->contents;

    if(index < 0) 
        {
        index = listlen(list) + index;
	if(index < 0) 
            {
            index = 0;
            insert = INSERT_BEFORE;
            }
	else insert = INSERT_AFTER;
        }
     else insert = INSERT_BEFORE;

    while(index--) 
        {
        if(list->next == nilCell)
		{
		insert = INSERT_END;
		break;
		}
	cell = list;
        list = list->next;
        }

    if(params == nilCell || !isList(list->type))  break;
    params = getIntegerExt(params, (UINT*)&index, evalFlag);
    }

newCell = copyCell(newCell);
if(insert == INSERT_BEFORE || list == nilCell)
    {
    if(list == (CELL*)cell->contents)
	{
	cell->contents = (UINT)newCell;
	newCell->next = list;
	}
    else
	{
	cell->next = newCell;
	newCell->next = list;
	}
    }

else if(insert == INSERT_AFTER || insert == INSERT_END)
    {
    cell = list->next;
    list->next = newCell;
    newCell->next = cell;
    }

return(copyCell(newCell));
}


CELL * p_pop(CELL * params)
{
CELL * list;
CELL * cell = NULL;
int index, evalFlag = FALSE;

list = evalCheckProtected(params, NULL);

if(!isList(list->type))
	{
	if(list->type == CELL_STRING)
		return(popString(list, params->next));
	else
		return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params));
	}

/* leave push optimization if popping first for queues */
if(params->next == nilCell)
	{
	cell = (CELL *)list->contents;
	list->contents = (UINT)cell->next;
	cell->next = nilCell;
	return(cell);
	}
else
	{
	list->aux = (UINT)nilCell; /* undo push optimization */
	cell = ((CELL*)params->next)->next;
	params = evaluateExpression(params->next);
	if(isList(params->type))
		{
		evalFlag = FALSE;
		params = getIntegerExt((CELL*)params->contents, (UINT*)&index, FALSE);
		}
	else 
		{
		evalFlag = TRUE;
		getIntegerExt(params, (UINT*)&index, FALSE);
		params = cell;
		}
	}

while(isList(list->type))
    {
    cell = list;
    list = (CELL *)list->contents;

    if(index < 0) index = convertNegativeOffset(index, list);

    while(index--) 
        {
        if(list->next == nilCell) break;
	cell = list;
        list = list->next;
        }

    if(params == nilCell || !isList(list->type))  break;
    params = getIntegerExt(params, (UINT*)&index, evalFlag);
    }

if(list == (CELL*)cell->contents)
    cell->contents = (UINT)list->next;
else
    cell->next = list->next;

list->next = nilCell;
return(list);
}


CELL * setNthStr(CELL * cellStr, CELL * new, int index, int typeFlag)
{
char * newStr;
char * oldStr;
UINT newLen, oldLen, len;
char * str;

oldStr = (char*)cellStr->contents;
oldLen = cellStr->aux - 1;

if(oldLen == 0) return(copyCell(cellStr));

#ifndef SUPPORT_UTF8

index = adjustNegativeIndex(index, oldLen);
deleteList((CELL*)sysSymbol[0]->contents);
sysSymbol[0]->contents = (UINT)stuffStringN(oldStr + index, 1);
len = 1;

#else

index = adjustNegativeIndex(index, utf8_wlen((char *)cellStr->contents));
str = oldStr;

while(index--) 
	{
	len = utf8_1st_len(str);
	str += len;
	}
len = utf8_1st_len(str);

deleteList((CELL*)sysSymbol[0]->contents);
sysSymbol[0]->contents = (UINT)stuffStringN(str, len);
index = str - oldStr;

#endif

getStringSize(new, &newStr, &newLen, TRUE);
/* get back oldStr in case it changed during eval of replacement */
oldStr = (char *)cellStr->contents;
oldLen = cellStr->aux - 1;
if(oldLen == 0) return(copyCell(cellStr));
index = adjustNegativeIndex(index, oldLen);

str = allocMemory(oldLen + newLen - len + 1);
*(str + oldLen + newLen - len) = 0;

memcpy(str, oldStr, index);
memcpy(str + index, newStr, newLen);
memcpy(str + index + newLen, oldStr + index + len, oldLen - index - len);

cellStr->contents = (UINT)str;
cellStr->aux = oldLen + newLen - len + 1;

if(typeFlag) 
	{
	new = stuffStringN(oldStr + index, len);
	freeMemory(oldStr);
	return(new);
	}

freeMemory(oldStr);
return(copyCell(cellStr));
}


CELL * popString(CELL * str, CELL * params)
{
char * ptr;
char * newPtr;
int index = 0;
int len = 1;
CELL * result;

if(params != nilCell)
	{
	params = getInteger(params, (UINT*)&index);
	if(params != nilCell) getInteger(params, (UINT*)&len);
	}

ptr = (char *)str->contents;

#ifndef SUPPORT_UTF8
index = adjustNegativeIndex(index, str->aux - 1);
#else
index = adjustNegativeIndex(index, utf8_wlen(ptr));
#endif

if(index + len > str->aux - 1 - index)
	len = str->aux - 1 - index;

newPtr = callocMemory(str->aux - len);

memcpy(newPtr, ptr, index);
memcpy(newPtr + index, ptr + index + len, str->aux - len - index);
str->aux = str->aux - len;
str->contents = (UINT)newPtr;
result = stuffStringN(ptr + index, len);
free(ptr);
return(result);
}


CELL * pushOnString(CELL * newStr, CELL * str, CELL * idx)
{
int index = 0;
char * ptr;
char * newPtr;
int minusFlag;

if(idx != nilCell) getInteger(idx, (UINT*)&index);
ptr = (char *)str->contents;

if(newStr->type != CELL_STRING)
	return(errorProcExt(ERR_STRING_EXPECTED, newStr));

if(index == -1)
	{
	appendCellString(str, (char *)newStr->contents, newStr->aux - 1);
	return(copyCell(newStr));
	}

minusFlag = (index < 0);

#ifndef SUPPORT_UTF8
index = adjustNegativeIndex(index, str->aux - 1);
#else
index = adjustNegativeIndex(index, utf8_wlen(ptr));
#endif

if(minusFlag && index) index++;

newPtr = allocMemory(str->aux + newStr->aux - 1);
memcpy(newPtr, ptr, index);
memcpy(newPtr + index, (char*)newStr->contents, newStr->aux - 1);
memcpy(newPtr + index + newStr->aux - 1, ptr + index, str->aux - index);
str->contents = (UINT)newPtr;
str->aux = str->aux + newStr->aux - 1;
*(newPtr + str->aux - 1) = 0;
free(ptr);

return(copyCell(newStr));
}


CELL * p_select(CELL * params)
{
UINT n = 0, idx  = 0; 
int index;
CELL * list, * cell;
CELL * result = NULL;
CELL * head;
int evalFlag = TRUE;
char * str, * newStr;
#ifdef SUPPORT_UTF8
int * wstr;
int * wnewStr;
int len;
#endif

head = evaluateExpression(params);
params = params->next;
cell = evaluateExpression(params);
if(isList(cell->type))
	{
	evalFlag = FALSE;
	cell = params = (CELL *)cell->contents;
	}

if(head->type == CELL_STRING)
	{
	if((n = listlen(params)) == 0) return(stuffString(""));
	
	str = (char *)head->contents;
#ifndef SUPPORT_UTF8
	newStr = (char *)allocMemory(n + 1);
	idx = 0;
	while(params->type != CELL_NIL)
		{
		if(idx == 0)
			{
			getIntegerExt(cell, (UINT *)&index, FALSE);
			params = params->next;
			}
		else
			params = getIntegerExt(params, (UINT *)&index, evalFlag);
		index = adjustNegativeIndex(index, head->aux -1);
		*(newStr + idx++) = *(str + index);
		}
	*(newStr + n) = 0;
#else
	wstr = allocMemory(head->aux * sizeof(int));
	len = utf8_wstr(wstr, str, head->aux - 1);
	wnewStr = allocMemory((n + 1) * sizeof(int));
	idx = 0;
	while(params->type != CELL_NIL)
		{
		if(idx == 0)
			{
			getIntegerExt(cell, (UINT *)&index, FALSE);
			params = params->next;
			}
		else
			params = getIntegerExt(params, (UINT *)&index, evalFlag);
		index = adjustNegativeIndex(index, len);
		*(wnewStr + idx++) = *(wstr + index);
		}
	*(wnewStr + n) = 0;
	newStr = allocMemory(UTF8_MAX_BYTES * n + 1);
	n = wstr_utf8(newStr, wnewStr, UTF8_MAX_BYTES * n);
	newStr = realloc(newStr, n + 1);
#endif
	result = getCell(CELL_STRING);
	result->aux = n + 1;
	result->contents = (UINT)newStr;	
	return(result);
	}

if(!isList(head->type))
	return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, head));
head = (CELL *)head->contents;
list = head;
n = 0;
while(params->type != CELL_NIL)
	{
	if(n++ == 0)
		{
		getIntegerExt(cell, (UINT *)&index, FALSE);
		params = params->next;
		}
	else
		params = getIntegerExt(params, (UINT *)&index, evalFlag);
	if(index < 0) index = convertNegativeOffset(index, head);
	if(index < idx) list = head, idx = 0;
	while(idx < index  && list->next != nilCell) list = list->next, idx++; 
	if(result == NULL)
		{
		result = getCell(CELL_EXPRESSION);
		cell = copyCell(list);
		result->contents = (UINT)cell;
		}
	else
		{
		cell->next = copyCell(list);
		cell = cell->next;
		}
	}

return((result == NULL) ? getCell(CELL_EXPRESSION) : result);
}	


CELL * p_slice(CELL * params)
{
CELL * cell;
UINT offset, length;

cell = evaluateExpression(params);
params = getInteger(params->next, &offset);
if(params != nilCell)
	getInteger(params, &length);
else
	length = (UINT)-1;
	
if(isList(cell->type))
        return(sublist((CELL *)cell->contents, (int)offset, length));

if(cell->type == CELL_STRING)
	return(substring((char *)cell->contents, cell->aux - 1, offset, length));

return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params));
}


CELL * sublist(CELL * list, int offset, UINT length)
{
CELL * subList;
CELL * cell;

if(offset < 0) 
	offset = convertNegativeOffset(offset, list);

subList = getCell(CELL_EXPRESSION);
if(length == 0) return(subList);

while(offset-- && list != nilCell)
	list = list->next;

if(list == nilCell) return(subList);

cell = copyCell(list);
subList->contents = (UINT)cell;
--length;
while(length--) 
	{
	list = list->next;
	if(list == nilCell) break;
	cell->next = copyCell(list);
	cell = cell->next;
	}

return(subList);
}


CELL * p_reverse(CELL * params)
{
CELL * cell;
CELL * previous;
CELL * next;
char * str;
int len, tmp;
char * left;
char * right;

cell = params;
params = evalCheckProtected(params, NULL);


if(isList(params->type))
	{
	params->aux = (UINT)nilCell; /* undo push optimization */

	previous = cell = (CELL*)params->contents;
	next = cell->next;
	cell->next = nilCell;
	while(cell!= nilCell)
		{
		previous = cell;
		cell = next;
		next = cell->next;
		if(cell != nilCell) cell->next = previous;
		}
	params->contents = (UINT)previous;
	}

else if(params->type == CELL_STRING)
	{
	str = (char *)params->contents;
	len = params->aux - 1;
	left = str;
	right = left + len - 1;
	while(left < right)
		{
		tmp = *left;
		*left = *right;
		*right = tmp;
		left++;
		right--;
		}
	}
else return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, cell));

return(copyCell(params));
}


CELL * p_join(CELL * params)
{
char * joint;
CELL * list;
int jointLen=0;

params = getListHead(params, &list);
if(params != nilCell)
	getStringSize(params, &joint, (UINT *)&jointLen, TRUE);
else
	joint = "";

return(append(NULL, list, joint, jointLen, FALSE));
}


CELL * p_find(CELL * params) 
{
char * key;
char * second;
UINT found;
CELL * next;
CELL * keyCell;
int size, options;

keyCell = evaluateExpression(params);
params = params->next;
next = evaluateExpression(params);

if(keyCell->type == CELL_STRING && next->type == CELL_STRING)
	{
	key = (char *)keyCell->contents;
	second = (char *)next->contents;
        size = next->aux - 1;

	if(params->next != nilCell)
            {
            params = getInteger(params->next, (UINT*)&options);
            found = searchBufferRegex(second, 0, key, size, options, NULL);
            if(found == 0xFFFFFFFF) return(nilCell);
            }
        else
            {
            found = searchBuffer(second, size, key, keyCell->aux - 1, TRUE);
            if(found == 0xFFFFFFFF) return(nilCell);
            }
	}
else
	{
	if(!isList(next->type)) return(nilCell);
	next = (CELL *)next->contents;
	found = 0;

	/* do regex when first arg is string and option# is present */
	if(keyCell->type == CELL_STRING && params->next != nilCell)
            {
	    getInteger(params->next, (UINT*)&options);
	    key = (char *)keyCell->contents;
	    while(next != nilCell)
		{
		if(next->type == CELL_STRING)
		    {
		    second = (char *)next->contents;
                    if(searchBufferRegex(second, 0, key, next->aux - 1 , options, NULL) != 0xFFFFFFFF)
		        break;
		    }
		found++;
		next = next->next;
		}
	    if(next == nilCell) return(nilCell);
            }
	else
	    {	
	    while(next != nilCell)
		{
		if(compareCells(keyCell, next) == 0) break;
		found++;
		next = next->next;
		}
	    if(next == nilCell) return(nilCell);
	    }
	}

return(stuffInteger(found));
}


void swap(UINT * left, UINT * right)
{
UINT tmp;

tmp = *left;
*left = *right;
*right = tmp;
}


CELL * p_swap(CELL * params)
{
int first, second, num;
char * str;
CELL * envelope;
CELL * list;
CELL * firstCell;
CELL * secondCell;

params = getInteger(params, (UINT*)&first);
params = getInteger(params, (UINT*)&second);

envelope = evalCheckProtected(params, NULL);

if(envelope->type == CELL_STRING)
	{
	first = adjustNegativeIndex(first, envelope->aux - 1);
	second = adjustNegativeIndex(second, envelope->aux - 1);
	str = (char *)envelope->contents;
	num = str[first];
	str[first] = str[second];
	str[second] = num;
	return(copyCell(envelope));
	}

if(!isList(envelope->type))
	return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params));

envelope->aux = (UINT)nilCell; /* undo push optimization */

list = (CELL *)envelope->contents;

if(first < 0) first = convertNegativeOffset(first, list);
if(second < 0) second = convertNegativeOffset(second, list);

if(first > second) swap((UINT*)&first, (UINT*)&second);
second = second - first;
	
firstCell = list;
while(first--)
	{
	if(firstCell->next == nilCell) break;
	firstCell = firstCell->next;
	}	
secondCell = firstCell;

while(second--)
	{
	if(secondCell->next == nilCell) break;
	secondCell = secondCell->next;
	}

swap(&firstCell->type, &secondCell->type);
swap(&firstCell->contents, &secondCell->contents);
swap(&firstCell->aux, &secondCell->aux);

return(copyCell(envelope));
}


CELL * p_dup(CELL * params)
{
CELL * list;
CELL * expr;
char * str;
UINT n, len;

expr = evaluateExpression(params);
params = params->next;
getInteger(params, &n);
if(n > 2147483647.0) n = 0;

if(expr->type == CELL_STRING && (evaluateExpression(params->next)->type == CELL_NIL) )
	{
	len = expr->aux - 1;
	list = getCell(CELL_STRING);
	str = allocMemory(len * n + 1);
	list->contents = (UINT)str;
	list->aux = (len * n + 1);
	*(str + len * n) = 0;
	while(n--) 
		{
		memcpy(str, (char *)expr->contents, len);
		str += len;
		}
	return(list);
	}

list = getCell(CELL_EXPRESSION);
if(n-- > 0) 
	{
	list->contents = (UINT)copyCell(expr);

	params = (CELL *)list->contents;
	while(n--)
		{
		params->next = copyCell(expr);
		params = params->next;
		}
	}

return(list);
}


#define STARTS_WITH 0
#define ENDS_WITH 1

CELL * startsEndsWith(CELL * params, int type)
{
char * string;
char * key;
int caseFlag=TRUE;
int klen, slen;
CELL * cell, * list;

cell = params->next;
list = evaluateExpression(params);
if(list->type == CELL_STRING)
    {
    string = (char *)list->contents;
    getString(cell, &key);
    }
else
    {
    if(!isList(list->type))
        errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params);
    cell = evaluateExpression(cell);
    list = (CELL *)list->contents;
   
    if(type == ENDS_WITH)
        while(list->next != nilCell) list = list->next;

    if(compareCells(list, cell) == 0) return(trueCell);
    else return(nilCell);
    }

if(cell->next != nilCell)
    caseFlag = (evaluateExpression(cell->next)->type != CELL_NIL);

if(type == STARTS_WITH)
	{
	if(caseFlag == TRUE) /* default*/
		{
		if(strncmp(string, key, strlen(key)) == 0)
			return(trueCell);
		}
	else  
		{
		if(my_strnicmp(string, key, strlen(key)) == 0)
			return(trueCell);
		}
	return(nilCell);
	}

klen = strlen(key);
slen = strlen(string);

if(klen > slen) return(nilCell);

if(caseFlag == TRUE) /* default */
	{
	if(strncmp(string + slen - klen, key, klen) == 0)
		return(trueCell);
	}
else
	{
	if(my_strnicmp(string + slen - klen, key, klen) == 0)
		return(trueCell);
	}

return(nilCell);
}

CELL * p_startsWith(CELL * params) { return startsEndsWith(params, STARTS_WITH); }
CELL * p_endsWith(CELL * params) { return startsEndsWith(params, ENDS_WITH); }

CELL * p_replace(CELL * params)
{
CELL * keyCell;
CELL * repCell;
CELL * list;
CELL * cell;
CELL * newList;
char * keyStr;
char * buff;
char * newBuff;
UINT cnt, newLen;
int options;

keyCell = evaluateExpression(params);
params = params->next;

newList = cell = evalCheckProtected(params, NULL);


cnt = 0;
if(isList(cell->type))
	{
	cell->aux = (UINT)nilCell; /* undo push optimization */

	list = (CELL *)cell->contents;

	if(params->next != nilCell)
		repCell = evaluateExpression(params->next);
	else
		repCell = NULL;
COMPARE_START:
	if(compareCells(keyCell, list) == 0)
		{
		if(repCell != NULL)
			{
			cell->contents = (UINT)copyCell(repCell);
			cell = (CELL*)cell->contents;
			cell->next = list->next;
			}
		else /* remove mode */
			cell->contents = (UINT)list->next;

		list->next = nilCell;
		deleteList(list);
		cnt++;

		if(repCell != NULL)
			list = cell;
		else /* remove mode */
			{
			list = (CELL*)cell->contents;
			goto COMPARE_START;
			}		
		}
	
	while(list->next != nilCell)
		{
		if(compareCells(keyCell, list->next) == 0)
			{
			cell = list->next;	/* cell = old elmnt */
			if(repCell != NULL)
				{
				list->next = copyCell(repCell);
				list = list->next;
				}
			list->next = cell->next;
			cell->next = nilCell;
			deleteList(cell);
			cnt++;
			}		
		else	
			list = list->next;
		}

	deleteList((CELL*)sysSymbol[0]->contents);	
	sysSymbol[0]->contents = (UINT)stuffInteger(cnt);
	return(copyCell(newList));
	}

if(cell->type == CELL_STRING)
	{
	if(keyCell->type != CELL_STRING)
		return(errorProc(ERR_STRING_EXPECTED));
	keyStr = (char *)keyCell->contents;
	buff = (char *)cell->contents;
	repCell = params->next;

	if(repCell == nilCell)
		return(errorProc(ERR_MISSING_ARGUMENT));
			
	options = -1;
	if(repCell->next != nilCell)
            getInteger(repCell->next, (UINT*)&options);

	newBuff = replaceString(keyStr, keyCell->aux - 1, 
	                       buff, cell->aux -1, repCell, &cnt, options, &newLen);
	if(newBuff != NULL)
	    {
	    freeMemory(buff);
	    cell->contents = (UINT)newBuff;
	    cell->aux = newLen + 1;
	    }

	deleteList((CELL*)sysSymbol[0]->contents);	
	sysSymbol[0]->contents = (UINT)stuffInteger(cnt);
	return(copyCell(cell));
	}

return(errorProcExt(ERR_LIST_OR_STRING_EXPECTED, params));
}


int adjustCount(int count, int length)
{
if(length == 0 || count == 0 || length == count)
	return(0);

if(count > 0)
    count = count % length;
else
    {
    count = -count % length;
    count = length - count;
    }
return(count);
}

CELL * p_rotate(CELL * params)
{
CELL * cell;
CELL * previous;
CELL * last = NULL;
int length, index;
int count;

cell = params;

if(cell->next != nilCell) getInteger(cell->next, (UINT *)&count);
else count = 1;

params = evalCheckProtected(params, NULL);

if(params->type == CELL_STRING)
	{	
	cell = copyCell(params);	
	length = params->aux - 1;
	if((count = adjustCount(count, length)) == 0) return(cell);
	memcpy((char*)cell->contents, (char *)(params->contents + length - count), count);
	memcpy((char*)(cell->contents + count), (char *)params->contents, length - count);
	memcpy((char*)params->contents, (char*)cell->contents, length);
	return(cell);
	}	

if(!isList(params->type))
	return(errorProcExt(ERR_LIST_EXPECTED, cell));

params->aux = (UINT)nilCell; /* undo push optimization */

cell = (CELL *)params->contents;
length = 0;
while(cell != nilCell)
	{
	++length;
	last = cell;
	cell = cell->next;
	}

if((count = adjustCount(count, length))== 0) return(copyCell(params));
index = length - count;

previous = cell = (CELL *)params->contents;
while(index--) 
	{
	previous = cell;
	cell = cell->next;
	}

previous->next = nilCell;
last->next = (CELL *)params->contents;
params->contents = (UINT)cell;

return(copyCell(params));
}

/* eof */
