/* nl-matrix.c --- matrix functions for newLISP
//
// Copyright (C) 1992-2006 Lutz Mueller <lutz@nuevatec.com>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2, 1991,
// as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
*/


#include "newlisp.h"
#include "protos.h"

#define TINY 1.0e-20


double * * transpose(double * * A, int m, int n);
double * * multiply(double ** A, double ** B, int m, int n, int k, int l);
double * * invert(double * * A, int n);
int ludcmp(double * * a, int n, int * indx, double * d);
void lubksb(double * * a, int n, int * indx, double * b);
int getDimensions(CELL * mat, int * n, int * m);
double * * list2matrix(CELL * list, int n, int m);
CELL * matrix2list(double * * matrix, int n, int m);
double * * allocateMatrix(int rows, int cols);
void freeMatrix(double * * m, int rows);

/* since version 7.4.8 'transpose' tramspose any matrix not only
   matrices containing numbers.
*/ 

CELL * p_matTranspose(CELL * params)
{
CELL * A;
CELL * B;
CELL * rowA = NULL;
CELL * cell = NULL;
CELL * new;
int n, m, i, j;
CELL * * Brows;

A = evaluateExpression(params);

if(A->type != CELL_EXPRESSION)
	return(errorProcExt(ERR_NOT_MATRIX, params));

if(getDimensions(A, &n, &m) == FALSE)
    return(errorProcExt(ERR_NOT_MATRIX, params));
    
if(n == 0 || m == 0)
    return(errorProcExt(ERR_WRONG_DIMENSIONS, params));

Brows = allocMemory(sizeof(CELL *) * m);
for(j = 0; j < m; j++)
	{
	Brows[j] = getCell(CELL_EXPRESSION);
	if(j > 0) Brows[j-1]->next = Brows[j];
	}

B = getCell(CELL_EXPRESSION);
B->contents = (UINT)Brows[0];

for(i = 0; i < n; i++)
	{
	if(i == 0)
		rowA = (CELL*)A->contents;
	else
		rowA = rowA->next;

	if(rowA->type != CELL_EXPRESSION)
		{
		free(Brows);
    		return(errorProcExt(ERR_LIST_EXPECTED, rowA));
		}

	for(j = 0; j < m; j++)
		{
		if(j == 0)
			cell = (CELL*)rowA->contents;
		else
			cell = cell->next;

		new = copyCell(cell);

		if( i == 0)
			Brows[j]->contents = (UINT)new;
		else
			Brows[j]->next = new;

		Brows[j] = new;
		}
	}

free(Brows);

return(B);
}



CELL * p_matMultiply(CELL * params)
{
CELL * A;
CELL * B;
CELL * C;
double * * X;
double * * Y;
double * * M;
int n, m, k, l;

B = params->next;
A = evaluateExpression(params);
B = evaluateExpression(B);

if(A->type != CELL_EXPRESSION)
	return(errorProcExt(ERR_NOT_MATRIX, params));
if(B->type != CELL_EXPRESSION)
	return(errorProcExt(ERR_NOT_MATRIX, params->next));

if(getDimensions(A, &n, &m) == FALSE)
	return(nilCell);
if(getDimensions(B, &k, &l) == FALSE)
	return(nilCell);

if((X = list2matrix(A, n, m)) == NULL)
	return(nilCell);

if((Y = list2matrix(B, k, l)) == NULL)
	{
	freeMatrix(X, n);
	return(nilCell);
	}

if((M = multiply(X, Y, n, m, k, l)) == NULL)
	{
	freeMatrix(X, n);
	freeMatrix(Y, k);
	return(nilCell);
	}

C = matrix2list(M, n, l);
freeMatrix(X, n);
freeMatrix(Y, k);
freeMatrix(M, n);

return(C);
}


CELL * p_matInvert(CELL * params)
{
CELL * dataA;
CELL * dataY;
int n, m;
double * * A;
double * * Y;

dataA = evaluateExpression(params);
if(dataA->type != CELL_EXPRESSION)
	return(errorProcExt(ERR_NOT_MATRIX, params));

if(getDimensions(dataA, &n, &m) == FALSE)
	return(nilCell);

if(n != m) return(errorProcExt(ERR_WRONG_DIMENSIONS, params));

if((A = list2matrix(dataA, n, n)) == NULL)
	return(nilCell);

if((Y = invert(A, n)) == NULL)
	{
	freeMatrix(A, n);
	return(nilCell);
	}

dataY = matrix2list(Y, n, n);

freeMatrix(A, n);
freeMatrix(Y, n);

return(dataY);
}


/* ------------------- C = A*B, A and B unchanged --------------------- */
double * * multiply(double ** A, double ** B, int n, int m, int k, int l)
{
double * * C;
int i, j, s;
double sum;

if((C = allocateMatrix(n, l)) == NULL)
	return(NULL);

if(m != k)
	{
	errorProc(ERR_WRONG_DIMENSIONS);
	return(NULL);
	}

for(i = 1; i <= n; i++)
	{
	for(j = 1; j <= l; j++)
		{
		sum = 0.0;
		for(s = 1; s <= m; s++)
			sum += A[i][s] * B[s][j];

		C[i][j] = sum;
		}
	}

return(C);
}


/* ----- return inverse of A in Y, A will contain LU decomposition --- */
double * * invert(double * * A, int n)
{
double * * Y;
double * col;
double d;
int i, j;
int * indx;

if((Y = allocateMatrix(n, n)) == NULL)
	return(NULL);

col = (double *)calloc(n + 4, sizeof(double));
indx = (int *)calloc((n + 1), sizeof(int));

if( ludcmp(A, n, indx, &d) != TRUE)
	return(NULL);

for(j = 1; j <= n; j++)
	{
	for(i = 1; i <= n; i++) col[i] = 0.0;
	col[j] = 1.0;
	lubksb(A, n, indx, col);
	for(i = 1; i <= n; i++) Y[i][j] = col[i];
	}

free(col);
free(indx);

return(Y);
}


/* ------------------------- LU decomposition --------------------------- */

int ludcmp(double * * a, int n, int * indx, double * d)
{
int i, imax = 0, j, k;
double big, dum, sum, temp;
double * vv;

vv = (double *)calloc(n + 4, sizeof(double));
*d = 1.0;

/* find abs biggest number in each row and put 1/big in vector */
for (i = 1; i <= n; i++)
	{
	big = 0.0;
	for(j = 1;j <= n; j++)
		if ((temp = fabs(a[i][j])) > big) big = temp;

	if (big == 0.0)
		{
		free(vv);
		errorProc(ERR_SINGULAR);
		return(FALSE);
		}

	vv[i] = 1.0 / big;
	}

for (j = 1; j <= n; j++)
	{
	for (i = 1; i < j; i++)
		{
		sum = a[i][j];
		for(k = 1; k < i; k++) sum -= a[i][k] * a[k][j];
		a[i][j] = sum;
		}
	big = 0.0;

	for (i = j; i <= n; i++)
		{
		sum = a[i][j];
		for(k = 1; k < j; k++)
			sum -= a[i][k] * a[k][j];
		a[i][j] = sum;

		if ( (dum = vv[i] * fabs(sum)) >= big)
			{
			big = dum;
			imax = i;
			}
		}

	if (j != imax)
		{
		for (k = 1; k <= n; k++)
			{
			dum = a[imax][k];
			a[imax][k] = a[j][k];
			a[j][k] = dum;
			}

		*d = -(*d);
		vv[imax] = vv[j];
		}


	indx[j] = imax;
	if (a[j][j] == 0.0) a[j][j] = TINY;

	if (j != n)
		{
		dum = 1.0 / (a[j][j]);
		for(i = j+1; i <= n; i++) a[i][j] *= dum;
		}
	}

free(vv);
return(TRUE);
}


void lubksb(double * * a, int n, int * indx, double * b)
{
int i, ii = 0, ip, j;
double sum;

for (i = 1; i <= n; i++)
	{
	ip = indx[i];
	sum = b[ip];
	b[ip] = b[i];
	if (ii)
		for(j = ii; j <= i-1; j++) sum -= a[i][j] * b[j];
	else
		if(sum) ii = i;
	b[i] = sum;
	}

for (i = n; i >= 1; i--)
	{
	sum = b[i];
	for(j = i+1; j <= n; j++) sum -= a[i][j] * b[j];
	b[i] = sum / a[i][i];
	}
}


int getDimensions(CELL * mat, int * n, int * m)
{
CELL * row;
CELL * cell;
int rows, cols;

row = (CELL *)mat->contents;
if(row->type != CELL_EXPRESSION)
	{
	errorProcExt(ERR_NOT_MATRIX, mat);
	return(FALSE);
	}

cell = (CELL *)row->contents;

rows = cols = 0;
while(row != nilCell)
	{
	++rows;
	row = row->next;
	}

while(cell != nilCell)
	{
	++cols;
	cell = cell->next;
	}

*n = rows;
*m = cols;

return(TRUE);
}

/* ------------ copy array/list into a matrix of doubles -------- */
double * * list2matrix(CELL * list, int n, int m)
{
double ** matrix;
CELL * row;
CELL * cell;
int i, j;
double fnum;

if((matrix = allocateMatrix(n, m)) == NULL)
	return(NULL);

row = (CELL *)list->contents;
i = 1;

while(row != nilCell && i <= n)
	{
	j = 1;
	cell = (CELL *)row->contents;
	while(cell != nilCell && j <= m)
		{
		if(cell->type == CELL_INT32) fnum = (int)cell->contents;
		else if(cell->type == CELL_FLOAT)
			memcpy(&fnum, (double *)&cell->aux, sizeof(double));
		else fnum = 0.0;
		matrix[i][j++] = fnum;
		cell = cell->next;
		}
	++i;
	row = row->next;
	}

return(matrix);
}


/* ------- allocate an array/list and copy matrix into it ---- */
CELL * matrix2list(double ** matrix, int n, int m)
{
CELL * list;
CELL * row;
CELL * cell;
int i, j;
double fnum;

list = getCell(CELL_EXPRESSION);
row = getCell(CELL_EXPRESSION);
list->contents = (UINT)row;

for(i = 1; i <= n; i++)
	{
	cell = getCell(CELL_FLOAT);
	row->contents = (UINT)cell;
	for(j = 1; j <= m; j++)
		{
		fnum = matrix[i][j];
		memcpy((double *)&cell->aux, &fnum, sizeof(double));
		if(j == m) break;
		cell->next = getCell(CELL_FLOAT);
		cell = cell->next;
		}
	if(i == n) break;
	row->next = getCell(CELL_EXPRESSION);
	row = row->next;
	}

return(list);
}


/* ------------------------ allocate/free a matrix ----------------- */
double * * allocateMatrix(int rows, int cols)
{
int i;
double * * m;

m = (double **) calloc(rows + 1, sizeof(double *));
if (!m)
	{
	errorProc(ERR_NOT_ENOUGH_MEMORY);
	return(NULL);
	}

for(i = 1; i <= rows; i++)
	{
	m[i] = (double *)calloc(cols + 1, sizeof(double));
	if (!m[i])
		{
		errorProc(ERR_NOT_ENOUGH_MEMORY);
		return(NULL);
		}
	}

return(m);
}


void freeMatrix(double * * m, int rows)
{
int i;

for(i = 1; i <= rows; i++)
	free(m[i]);

free(m);
}

/* end of file */
